
// A simple demonstration of RAII (Resource Allocation Is Initialization)
//
//

#include <iostream>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>

std::string lsof_command;

class RAII_file {
    public:
        FILE* fh;
        RAII_file( const char* filename ) : fh(NULL) {
            if ( filename ) {
                fh = fopen( filename, "w" );
            }
        }
        ~RAII_file() {
            if ( fh ) {
                fclose( fh );
            }
        }
};

void do_file_stuff() {
    RAII_file F( "foo.txt" );
    fprintf( F.fh, "hello world\n" );

    std::cout << "\nopen files during call to do_file_stuff:" << std::endl;
    system( lsof_command.c_str() );

}


int main() {
    // Set up to run Unix lsof (list open files) for this process
    pid_t mypid = getpid();
    std::stringstream lsof;
    lsof << "/usr/sbin/lsof +d . -p " << mypid << " | grep foo.txt";
    lsof_command = lsof.str();
    std::cout << lsof_command << std::endl;
    
    std::cout << "\nopen files to start with:" << std::endl;
    system( lsof_command.c_str() );

    do_file_stuff();

    std::cout << "\nopen files at the end:" << std::endl;
    system( lsof_command.c_str() );


}

