

#include <tcl.h>
#include <iostream>
#include <sstream>
#include <vector>
#include <map>
#include <string>
#include <string.h>
#include <stdlib.h> 

using std::cout;
using std::endl;

// some kind of iterator
struct Foo {
    Foo( size_t tracker_id, size_t id ) : ref_count(0), tracker_index(tracker_id), index(id), tracker(NULL) {
        std::stringstream foo_name;
        foo_name << tracker_index << ":Foo:" << index;
        name = foo_name.str();
    }
    int ref_count;
    size_t tracker_index;
    size_t index;
    std::string name;
    struct FooTracker* tracker;
};

struct FooKeeper {
    std::vector<Foo*> Foos;
    static int last_index;
    int index;
    FooKeeper( int arg=0 ) : index(arg) {}
};

// probably Ext_Obj_Table or Ext_Hash_Map (ext_table.h)
static std::map<int,FooKeeper> AllFoos;
int FooKeeper::last_index = 0;

bool GetTrackerIndex( size_t &tracker_index, const char* name ) {
    char* end;
    tracker_index = strtol( name, &end, 10 );
    // corrupted name, can't find a tracker.
    if ( name == end ) return false;
    std::map<int,FooKeeper>::iterator tracker_entry = AllFoos.find(tracker_index);
    if ( tracker_entry == AllFoos.end() ) { 
        // Tracker is gone, can't find the Foo.
        return false; 
    }
    return true;
}

Foo* GetFooByName( const char* foo_name ) {
    size_t tracker_index;
    if ( !GetTrackerIndex( tracker_index, foo_name ) ) { return NULL; }
    for( size_t idx=0; idx != AllFoos[tracker_index].Foos.size(); ++idx ) {
        if ( AllFoos[tracker_index].Foos[idx] && foo_name == AllFoos[tracker_index].Foos[idx]->name ) return AllFoos[tracker_index].Foos[idx];
    }
    return NULL;
}

void Dup_Tcl_Foo_InternalRep(Tcl_Obj *srcPtr, Tcl_Obj *copyPtr);
void Free_Tcl_Foo_InternalRep(Tcl_Obj *g );
void UpdateStringOf_Tcl_Foo(Tcl_Obj *p);
Tcl_ObjType Tcl_Foo = {
    (char*)"Foo_Value",
    Free_Tcl_Foo_InternalRep,
    Dup_Tcl_Foo_InternalRep,
    UpdateStringOf_Tcl_Foo,
    NULL
};

void Free_Tcl_FooTracker_InternalRep(Tcl_Obj *g )
{
    FooKeeper* the_keeper = (FooKeeper*)g->internalRep.otherValuePtr;
    cout << "Free was called, cleaning up all locally tracked Foo's from Keeper:" << the_keeper->index << endl;
    for( std::map<int,FooKeeper>::iterator i = AllFoos.begin(); i != AllFoos.end(); ++i ) {
        for( size_t idx=0; idx != the_keeper->Foos.size(); ++idx ) {
            if ( the_keeper->Foos[idx] ) {
                cout << "Deleting Foo " << the_keeper->Foos[idx]->name << endl;
                delete the_keeper->Foos[idx];
            }
            the_keeper->Foos[idx] = NULL;
        }
    }
}

int Set_Tcl_FooTracker_FromAny(Tcl_Interp *interp , Tcl_Obj * the_foo)
{
    cout << "Set_FooTracker_FromAny was called" << endl;
    abort();
    return TCL_OK;
}

void Dup_Tcl_FooTracker_InternalRep(Tcl_Obj *srcPtr, Tcl_Obj *copyPtr)
{

     cout << "Dup_Tcl_FooTracker_InternalRep was called" << endl;
     copyPtr->typePtr = srcPtr->typePtr;

     // Point at the same item in the All
     Foo* f = reinterpret_cast<Foo *>(srcPtr->internalRep.otherValuePtr);
     copyPtr->internalRep.otherValuePtr = f;
     f->ref_count++;
}

void UpdateStringOf_Tcl_FooTracker(Tcl_Obj *p)
{
    cout << "UpdateStringOf FooTracker was called" << endl;
    FooKeeper* keeper = ((FooKeeper*)p->internalRep.otherValuePtr);
    std::stringstream foo_keeper_name;
    foo_keeper_name << "Keeper:";
    foo_keeper_name  << keeper->index;
    std::string foo_keeper_string = foo_keeper_name.str();
    p->length = strlen(foo_keeper_string.c_str());
    p->bytes = (char*)(ckalloc((unsigned) p->length+1 ));
    strcpy(p->bytes, foo_keeper_string.c_str());
    return;
}

Tcl_ObjType Tcl_Foo_tracker = {
    (char*)"Foo_Value_scoped_tracker",
    Free_Tcl_FooTracker_InternalRep,
    Dup_Tcl_FooTracker_InternalRep,
    UpdateStringOf_Tcl_FooTracker,
    NULL
};



void Free_Tcl_Foo_InternalRep(Tcl_Obj *g )
{
    // This doesn't free the internal representation unless it can not find its tracker.
    Foo* f = reinterpret_cast<Foo *>(g->internalRep.otherValuePtr);
    cout << "Free Foo was called on " << f->name << endl;
    if ( --f->ref_count == 0 ) {
        size_t tracker_index = f->tracker_index;
        // Make sure that the tracker is still around
        std::map<int,FooKeeper>::iterator tracker_entry = AllFoos.find(tracker_index);
        if ( tracker_entry != AllFoos.end() ) { 
            AllFoos[tracker_index].Foos[f->index] = NULL;
        }
        cout << "Foo " << f->name << " was deleted." << endl;
        delete f;
    }
}

int Set_Tcl_Foo_FromAny(Tcl_Interp *interp , Tcl_Obj * the_foo)
{
    cout << "Set_Tcl_Foo_FromAny was called" << endl;
    const char* foo_tracker;
    foo_tracker = Tcl_GetStringFromObj(the_foo, NULL);
    Foo* foo = GetFooByName( foo_tracker );
    if ( !foo ) {
        return TCL_ERROR;
    }
    cout << "Reconstituted " << foo->name << " from Keeper:" << foo_tracker << endl;
    foo->ref_count++;
    the_foo->internalRep.otherValuePtr = (void*)foo;
    the_foo->typePtr = &Tcl_Foo;
    return TCL_OK;
}

void Dup_Tcl_Foo_InternalRep(Tcl_Obj *srcPtr, Tcl_Obj *copyPtr)
{

     cout << "InternalRep was called" << endl;
     copyPtr->typePtr = srcPtr->typePtr;

     // Is this safe?
     copyPtr->internalRep.otherValuePtr = srcPtr->internalRep.otherValuePtr;
}

void UpdateStringOf_Tcl_Foo(Tcl_Obj *p)
{
    cout << "UpdateStringOf_Tcl_Foo was called" << endl;
    std::string& foo_name = ((Foo*)p->internalRep.otherValuePtr)->name;
    if ( !p->bytes ) {
        p->length = strlen(foo_name.c_str());
        p->bytes = (char*)(ckalloc((unsigned) p->length+1 ));
        strcpy(p->bytes, foo_name.c_str());
    }
    return;
}

const char* FooTrackerVarname = "FooTracker__secret_variable_name";
Tcl_Obj * Create_Tcl_Foo (Tcl_Interp* interp )
{
    size_t tracker_index = 0;
    FooKeeper* the_keeper = NULL;
    Tcl_Obj* local_foo_tracker = Tcl_GetVar2Ex( interp, FooTrackerVarname, NULL, 0);
    if ( !local_foo_tracker ) {
        // First time one is created at this scope
        tracker_index = FooKeeper::last_index++;
        AllFoos[tracker_index] = FooKeeper(tracker_index);
        the_keeper = &AllFoos[tracker_index];
        Tcl_Obj *new_obj = Tcl_NewObj();
        new_obj->typePtr = &Tcl_Foo_tracker;
        new_obj->internalRep.otherValuePtr = the_keeper;
        UpdateStringOf_Tcl_FooTracker(new_obj);

        Tcl_SetVar2Ex( interp, FooTrackerVarname, NULL, new_obj, 0 );
    } else if ( local_foo_tracker->typePtr == &Tcl_Foo_tracker ) {
        // Use the previously created tracker.
        the_keeper = (FooKeeper*)local_foo_tracker->internalRep.otherValuePtr;
        tracker_index = the_keeper->index;
    } else {
        cout << "FooKeeper was corrupted, shimmering may occur!" << endl;
    }
    size_t f_index = the_keeper->Foos.size();
    Foo* f = new Foo(tracker_index, f_index);
    f->ref_count = 1;
    Tcl_Obj *objPtr = Tcl_NewObj();

    objPtr->bytes = NULL;

    objPtr->internalRep.otherValuePtr = (void*)f;
    objPtr->typePtr = &Tcl_Foo;

    if ( the_keeper ) {
        the_keeper->Foos.push_back(f);
        f->ref_count++;
    }
    cout << "Created Foo: " << f->name << endl;
    return objPtr;
}



// some YS command
int do_Foo(
    ClientData c_data,
    Tcl_Interp *interp,
    int argc,
    Tcl_Obj *     CONST argv[]
)
{
    const char* usage = "usage: do_foo cmd [ foo_obj ]";
    if ( argc < 2 || argc > 4 ) {
        cout << usage << endl;
        return TCL_ERROR;
    }
    Tcl_Obj* the_foo = NULL;
    if ( argc == 3 ) {
        // second argument should be a foo
        the_foo = argv[2];
        if ( the_foo->typePtr != &Tcl_Foo ) {
            // Must have shimmered, see if we can reconstruct it
            // There seems to be some sort of barrier to calling this for a Unicode string.
            //if ( Tcl_ConvertToType( interp, the_foo, &Tcl_Foo ) != TCL_OK ) {
            //    cout << "Not a foo" << endl;
            //    return TCL_ERROR;
            //}
            if (the_foo->typePtr->freeIntRepProc) {
                the_foo->typePtr->freeIntRepProc(the_foo);
            }
            if ( Set_Tcl_Foo_FromAny(interp, the_foo) != TCL_OK ) {
                return TCL_ERROR;
            }
        }
    }
    std::string cmd = Tcl_GetStringFromObj( argv[1], NULL );
    if ( argc == 2 && cmd == "make" ) {
        the_foo = Create_Tcl_Foo(interp);
        Tcl_SetObjResult( interp, the_foo );
        return TCL_OK;
    } else if ( argc == 3 && cmd == "name" ) {
        Tcl_Obj* result = Tcl_NewStringObj( "Harry", -1 );
        Tcl_SetObjResult( interp, result );
        return TCL_OK;
    }
    cout << "Invalid argument or command for do_foo" << endl;
    cout << usage << endl;
    return TCL_ERROR;
}

int Foo_Appinit( Tcl_Interp* interp ) {
    if ((Tcl_Init)(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }

    // equivalent of any get_cells, get_data, etc...
    Tcl_CreateObjCommand(interp, "do_foo", do_Foo, NULL, NULL);
    return TCL_OK;
}

int main( int argc, char* argv[] ) {

    Tcl_Main( argc, argv, &Foo_Appinit );

}
